[comment {-*- tcl -*- doctools manpage $Id: scrobj.tcl,v 1.3 2008/12/22 10:53:06 cn Exp $}]
[manpage_begin scrobj n 1.0]
[moddesc   {scripted Tcl_Objects}]
[titledesc {scripted Tcl_Objects}]
[require Tcl 8.5]
[require scrobj 1.0]
[description]
[para]
Tcl's fundamental mantra says that "everything is a string" and
much of Tcl's simplicity derives from this rule. Strings can,
however, carry with them an internal representation as some other
data type, and this is one reason for Tcl's astonishing performance.
For example, Tcl maintains such an internal representation for integers,
lists, bytecodes, and many other types. 

[para]
Extension writers have long been able to add new custom types 
using Tcl's C interface. The scrobj package allows you to do the same
from the script level. 

[para]
Unfortunately, in order to achieve this the package has to
violate a basic assumption of Tcl's bytecode engine, namely that
it is not getting invoked recursively from a call to Tcl_GetString. 
A consequence is that scrobj can be used to crash the application,
although this does require a rather contrived setup. 

[para]
To prevent such crashes in the legitimate cases the package uses 
a dedicated Tcl interpreter for every registered type. Both
Tcl 8.5 and Tcl 8.6 seem to tolerate recursive invocations
of the bytecode engine, as long as they execute in different
interpreters.

[para]
[section "Usage"]

[list_begin definitions]

[call [cmd scrobj] [cmd register] [arg typeName] [arg updateCode] [arg parseCode]]

To register a new data type you have to write code 
to transform the string value to the internal representation 
and vice versa. This code then needs to be registered using 
[cmd scrobj] [cmd register].

[nl]
[example_begin]
scrobj register mytype {
    intRep {
	# code to generate the string
	# from the internal representation
	set dummy "<$intRep>"
    }
} {
    stringVal {
	# code to generate the internal 
	# representation from the string
	regexp -- [format "{^<(%c0-9%c+)>\$}" 91 93] $stringVal -> irp
	set irp
    }
}
[example_end]

[nl]
The call to [cmd scrobj] [cmd register] implicitly creates a new Tcl 
interpreter where the conversion routines will be executed.

[call [cmd scrobj] [cmd convert] [arg typeName] [arg string] [opt extraArg]]

Return the internal representation corresponding to [arg string]
for the type [arg typeName]. If the internal representation is
already available, that value will be returned directly. Otherwise 
the internal representation will first be generated by applying
the type's [arg parseCode] to the given [arg string]. 

[nl]
[example_begin]
# convert "<4711>" and return the internal representation "4711"
scrobj convert mytype <4711>
[example_end]

[call [cmd scrobj] [cmd value] [arg typeName] [arg intRep] [opt extraArg]]

Generate a value from the internal representation.

[nl]
[example_begin]
# generate a value for type mytype
set a [format "%cscrobj value mytype 123%c" 91 93]

# the next call will [strong not] invoke the 
# parseCode because [format %ca 36] already has an internal
# representation for type "mytype"
set b [format "%cscrobj convert mytype \$a%c" 91 93]

# the next line will run the updateCode for mytype
set c "a is $a" ;# -> "a is <123>"
[example_end]

[call [cmd scrobj] [cmd eval] [arg typeName] [arg code]]

Execute some [arg code] in the interpreter for type [arg typeName].

[call [cmd scrobj] [cmd info] [opt typeName]]

Return the list of all registered types, or information about
one specific type.

[list_end]

[section "Parametrized types"]

The [cmd convert] and [cmd value] subcommands allow to 
specify an optional extra argument. This can be used
to parametrize a data type by a value:

[proc myrep stuff {string map {< [ > ] # $} $stuff}]

[list_begin definitions]
[lst_item Example:]
[example_begin]
# register a parametrized type that uses a configurable
# regular expression for the conversion
scrobj register rexp {
    irep {
        # string is kept in internal rep
        lindex $irep 1
    }
} {
    {inString regExp} {
        regexp -- $regExp $inString -> res
        list $res $inString
    }
} 

set pat1 {[myrep ^foo(<0-9>+)$]}
set pat2 {[myrep ^foo(<0-9>+)$]}
set pat3 {[myrep ^fo(o<0-9>+)$]}

set a [myrep {<scrobj value rexp <scrobj convert rexp foo17 #pat1> #pat1>}]

# The next line will use the internal representation
# of $a that's already there
set b [myrep {<scrobj convert rexp #a #pat1>}] 

# The next line will re-generate the internal
# representation, because $pat1 and $pat2
# are not represented by the same Tcl_Object.
set c [myrep {<scrobj convert rexp #a #pat2>}]

# The next line also re-generates the internal
# representation. This time the return value
# is actually different.
set d [myrep {<scrobj convert rexp #a #pat3>}]
[example_end]
[list_end]
[section "Application examples"]

scrobj comes equipped with two packages, 
[emph scrobj::expression] and [emph scrobj::rational],
which illustrate the use of scrobj types.

[list_begin definitions]
[lst_item scrobj::expression]
[emph scrobj::expression] implements a parser for simple arithmetic 
expressions. 

[nl]
[example_begin]
package require scrobj::expression 1.0

scrobj convert scrobj::expression {23*foo(bar(12),3+$a/7)} {
    opns ::operatorns::
    funcns ::funcns::
}

# the result of the conversion is Tcl code that 
# implements the computation:
[myrep {{
    upvar 1 a 3 
    set 1 [::funcns::bar [eval {set dummy 12}]]
    set 2 [::funcns::foo [eval {set 1}] [eval {
	set 4 [::operatorns::/ $3 7]
	set 5 [::operatorns::+ 3 $4]
	set 5
    }]]
    set 6 [::operatorns::* 23 $2]
    set 6
} {23*foo(bar(12),3+$a/7)}}]
[example_end]

[nl]
The parser is implemented in Tcl using regular expressions.
It doesn't have to be fast, because the script
will be cached in the internal representation once it is
generated.

[lst_item scrobj::rational]
The [emph scrobj::rational] package implements rational number
arithmetic. A rational number p/q is represented internally as
the list [cmd "[lb]list \$p \$q[rb]"]. 

[nl]
[example_begin]
package require scrobj::rational 1.0

scrobj::rational numerator 17272/9128 ;# -> 17272
scrobj::rational + 1727/162 8348/1731 ;# -> 1447271/93474
[example_end]

[nl]
[emph scrobj::rational] uses [emph scrobj::expression] to implement
its own version of the [cmd expr] command:

[nl]
[example_begin]
package require scrobj::rational 1.0

# register some custom functions
namespace eval rational {
    namespace eval func {
	proc f1 {a b} {
	    ::scrobj::rational expression {3*$a-17/93*$b}
	}
	proc f2 {x} {
	    ::scrobj::rational expression {2*$x-1}
	}
    }
}

set a 72/841

scrobj::rational expression {15*f1($a,3-$a) - f2(17/18+$a)/9}
# -> -8988239/2111751
[example_end]

[list_end]

[keywords scrobj rational expression Tcl_Object]

[manpage_end] 