#
# A simple expression parser to illustrate the scrobj package 
#
# Copyright (c) 2008 Christian Nassau
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: expression.tcl,v 1.8 2008/12/03 15:34:27 cn Exp $
#


package require scrobj 1.0
package provide scrobj::expression 1.0

scrobj register scrobj::expression {
    irep {
	# we keep the original string as part of the internal representation
	lindex $irep 1
    }
} {
    {str {config {}}} {
	list [expression::parse $str $config] $str
    }
}

scrobj eval scrobj::expression {

    namespace eval expression {

	# A simple expression parser based on the regexp command - it does not 
	# have to be fast, since we're counting on the internal representation
	# to remain intact for most of the application's lifetime.
	
	if 0 {
	    
	    # The parser transforms its input into a Tcl script that can then 
	    # be eval'ed. For example "3 + 5*(7+$a) - myfunc ($a, 3*$c)" becomes 
	    # something like this:
	
	    upvar 1 a 2 a 5 c 6 
	    set 1 [eval {
		set 3 [+ 7 $2]
		set 3
	    }]
	    set 4 [ myfunc [eval {
		set 5
	    }] [eval {
		set 7 [* 3 $6]
		set 7
	    }]]
	    set 8 [* 5 $1]
	    set 9 [+ 3 $8]
	    set 10 [- $9 $4]
	    set 10
	}

	# regular expressions for the basic arithemtic operations
	foreach op {*/ +-} {
	    set regexps($op) [string map "%%OP%% $op" \
                                  {^(|[^%%OP%%]*[^$0-9%%OP%%])(\$?[0-9]+)\s*([%%OP%%])\s*(\$?[0-9]+)(|[^$0-9].*)$}]
	}

        proc FindParenOrFunction {exp -> av funcnamev subexprv bv} {
            upvar 1 $av a $funcnamev funcname $subexprv subexpr $bv b
            if {[regexp -- {^(|.*[^a-zA-Z])(\m[a-zA-Z0-9]*)\s*\(([^()]+)\)(.*)$} $exp -> a funcname subexpr b]} {
                return 1
            }
            if {[regexp -- {^(|.*[^a-zA-Z])\s*\(([^()]+)\)(.*)$} $exp -> a subexpr b]} {
                set funcname ""
                return 1
            }
            return 0
        }

	proc Parse {exp} {
	    variable varcnt
	    variable varlist
	    variable regexps
	    variable cfg
	    set bdy {}

            while 1 {

                # handle parentheses & function calls
                while {[FindParenOrFunction $exp -> a funcname subexpr b]} {
                    set myvar [incr varcnt]
                    set exp "$a \$$myvar $b"
                    if {$funcname eq ""} {
                        set call "set $myvar"
                    } elseif {$funcname eq "negate"} {
                        set call "set $myvar \[$cfg(opns)$funcname"
                    } else {
                        set call "set $myvar \[$cfg(funcns)$funcname"
                    }
                    set acnt 0
                    foreach arg [split $subexpr ,] {
                        append call " \[[list eval [Parse $arg]]\]"
                        incr acnt
                    }
                    if {$funcname eq ""} {
                        if {$acnt > 1} {error {unexpected "," outside function argument list}}
                    } else {
                        append call "\]"
                    }
                    lappend bdy $call
                }
                
                # replace variables 
                while {[regexp -- {^(.*)(\$)([a-zA-Z][a-zA-Z0-9]*)([^a-zA-Z0-9].*|)$} $exp -> a dollar varname b]} {
                    set myvar [incr varcnt]
                    lappend varlist $varname $varcnt 
                    set exp "$a \$$myvar $b"
                }

                # rewrite "... + -$5" as "... + negate($5)"
                if {[regexp -- {^(.*)([+*/-])\s*[-]\s*(\$?[0-9]+)([^0-9].*|)$} $exp -> a op varname b]} {
                    set exp "$a $op negate($varname) $b"
                    continue
                }

                # rewrite " -$5" as " negate($5)"
                if {[regexp -- {^(\s*)[-]\s*(\$?[0-9]+)([^0-9].*|)$} $exp -> a varname b]} {
                    set exp "$a negate($varname) $b"
                    continue
                }

                # rewrite "... + +$5" as "... + $5"
                if {[regexp -- {^(.*)([+*/-])\s*[+]\s*(\$?[0-9]+)([^0-9].*|)$} $exp -> a op varname b]} {
                    set exp "$a $op $varname $b"
                    continue
                }

                # rewrite " +$5" as " $5"
                if {[regexp -- {^(\s*)[+]\s*(\$?[0-9]+)([^0-9].*|)$} $exp -> a varname b]} {
                    set exp "$a $varname $b"
                    continue
                }

                foreach op {*/ +-} {
                    while {[regexp -- $regexps($op) $exp -> x a o b y]} {
                        set myvar [incr varcnt]
                        set exp "$x \$$myvar $y"
                        lappend bdy "set $myvar \[$cfg(opns)$o $a $b\]"
                    }
                }

                if {[regexp -- {^\s*(\$?)([0-9]+)\s*$} $exp -> dollar varname]} {
                    if {$dollar eq ""} {
                        lappend bdy "set dummy $varname"
                    } else {
                        lappend bdy "set $varname"
                    }
                } else {
                    error "component expression $exp not understood"
                }
                break
            }

	    join $bdy \n
	}

	proc parse {expr config} {
	    variable varlist {}
	    variable varcnt 0
	    variable cfg
	    array set cfg {
		funcns {}
		opns {}
	    }
	    array set cfg $config
	    set bdy [Parse $expr]
	    if {[llength $varlist]} {
		set bdy "[list upvar 1 {*}$varlist] \n $bdy"
	    }
	    return $bdy
	}

    }
}

# we can now implement our own version of the "expr" command
namespace eval scrobj {
    namespace eval "expr" {
                   
        namespace eval mathfunc {
            foreach {tclFunc argspec} {
                max {a b} min {a b} abs a bool a pow {a b} int a
            } {
                set x [string map {a $a b $b} $argspec]
                proc $tclFunc $argspec "::expr {$tclFunc ([join $x ,])}" 
            }
            proc negate a {::expr {-$a}}
      
	    foreach op {+ - * /} {
		proc $op {a b} "::expr {\$a $op \$b}"
	    }
	}

        # use a new scrobj type that inserts the mathfunc namespace

        scrobj register scrobj::expr2 {
            irep {
                list $irep 2
            }
        } {
            inString {
                set script [lindex [scrobj convert scrobj::expression $inString {
		    opns ::scrobj::expr::mathfunc::
		    funcns ::scrobj::expr::mathfunc::
		}] 0]
                list $script $inString
            }
        }

        proc expr {stuff} {
            eval [lindex [scrobj convert scrobj::expr2 $stuff] 0]
        }
    }
}
