#
# Rational number arithmetic to illustrate the scrobj package 
#
# Copyright (c) 2008 Christian Nassau
#
# See the file "license.terms" for information on usage and redistribution of
# this file, and for a DISCLAIMER OF ALL WARRANTIES.
#
# RCS: @(#) $Id: rational.tcl,v 1.3 2008/12/03 15:34:27 cn Exp $
#

package require scrobj 1.0
package require scrobj::expression 1.0
package provide scrobj::rational 1.0

# Register the object type for rational numbers. A fraction $a/$b 
# is represented as [list $a $b].

scrobj register scrobj::rational {
    irep {
	foreach {num den} $irep break
	# update string from internal rep
	if {$den != 1} {
	    return $num/$den
	} 
	return $num
    }
} {
    inp {
	# parse rational number from string
	if {[regexp {^([*-]?[0-9]+)(/[0-9]+)?$} $inp -> num den]} {
	    if {$den eq ""} {
		return [list $num 1]
	    }
	    set den [string range $den 1 end]
	    if {$den != 0} {
		return [list $num $den]
	    }
	    return -code error "denominator must not be zero"
	}
	return -code error "rational number expected"
    }
}

namespace eval scrobj::rational {
    
    proc numerator {value} {
	lindex [scrobj convert scrobj::rational $value] 0
    }
    proc denominator {value} {
	lindex [scrobj convert scrobj::rational $value] 1
    }

    namespace eval tcl::mathfunc {}
    namespace eval tcl::mathfunc::gcdhelper {}

    proc tcl::mathfunc::gcd {a b} {
	if {!$a} {return $b}
	if {!$b} {return $a}
	if {$a<0} {set a [expr {-$a}]}
	if {$b<0} {set b [expr {-$b}]}
	if {$a<$b} {
	    gcdhelper::gcd2 $b $a
	} else {
	    gcdhelper::gcd2 $a $b
	}
    }

    proc tcl::mathfunc::gcdhelper::gcd2 {a b} {
	# assume a >= b
	if {1 == $b} {return 1}
	set q [expr {$a/$b}]
	set r [expr {$a%$b}]
	if {0 == $r} {return $b}
	gcd2 $b $r
    }

    proc tcl::mathfunc::lcm {a b} {expr {$a*$b/gcd($a,$b)}}

    proc cancel {value} {
	foreach {num den} [scrobj convert scrobj::rational $value] break
	set gcd [expr {gcd($num,$den)}]
	scrobj value scrobj::rational [list [expr {$num/$gcd}] [expr {$den/$gcd}]]
    }

    proc + {a b} {
	foreach {num1 den1} [scrobj convert scrobj::rational $a] break
	foreach {num2 den2} [scrobj convert scrobj::rational $b] break
	cancel [scrobj value scrobj::rational [list [expr {$num1*$den2+$num2*$den1}] [expr {$den1*$den2}]]]
    }
    proc - {a b} {
	foreach {num1 den1} [scrobj convert scrobj::rational $a] break
	foreach {num2 den2} [scrobj convert scrobj::rational $b] break
	cancel [scrobj value scrobj::rational [list [expr {$num1*$den2-$num2*$den1}] [expr {$den1*$den2}]]]
    }
    proc * {a b} {
	foreach {num1 den1} [scrobj convert scrobj::rational $a] break
	foreach {num2 den2} [scrobj convert scrobj::rational $b] break
	cancel [scrobj value scrobj::rational [list [expr {$num1*$num2}] [expr {$den1*$den2}]]]
    }
    proc / {a b} {
	foreach {num1 den1} [scrobj convert scrobj::rational $a] break
	foreach {num2 den2} [scrobj convert scrobj::rational $b] break
	set d [expr {$num1*$den2}]
	set n [expr {$den1*$num2}]
	if {$n<0} {
	    set d [expr {-$d}]
	    set n [expr {-$n}]
	}
	cancel [scrobj value scrobj::rational [list $d $n]]
    }

    proc negate {a} {
	- 0 $a
    }

    proc != {a b} {
	expr {![== $a $b]}
    }

    proc == {a b} {
	iszero [- $a $b]
    }
    
    proc iszero {a} {
	expr {0 == [numerator $a]}
    }

    proc expression {exp} {
	upvar 1 __scrobj_rational_expression_cfg__ cfg
	if {![info exists cfg]} {
	    set callerns [uplevel 1 namespace current]
	    lappend cfg funcns ${callerns}::rational::func::
	}
        eval [lindex [scrobj convert scrobj::expression $exp $cfg] 0]
    }

    namespace export + - / * == != iszero numerator denominator cancel expression
    namespace ensemble create 
}
